
from typing import Optional, Any, Union, Callable
import torch as tt
import torch.nn as nn
from torch import Tensor
from torch.nn import functional as F

#from .modular import duplicate as _get_clones
__all__ = [
    'TransformerEncoder',       'TransformerDecoder',       'TransformerCross',
    'TransformerEncoderLayer',  'TransformerDecoderLayer',  'TransformerCrossLayer',
    'Encoder', 'Decoder',
    'CODER_DEFAULT_ARGS',
    ]


def _get_activation_fn(activation: str) -> Callable[[Tensor], Tensor]:
    if activation == "relu":
        return F.relu
    elif activation == "gelu":
        return F.gelu

    raise RuntimeError("activation should be relu/gelu, not {}".format(activation))

CODER_DEFAULT_ARGS = dict(
    d_model=            512, 
    nhead=              8, 
    dim_feedforward =   2048, 
    dropout =           0.1,
    activation=         F.gelu,
    normF =             nn.LayerNorm,
    normA =             dict(eps=1e-6),
    norm_first =        False,
    attention2F =       (nn.MultiheadAttention, dict(batch_first=True)),
)


class TransformerEncoderLayer(nn.Module):
    __constants__ = ['batch_first', 'norm_first']

    def __init__(self, d_model, nhead, dim_feedforward, dropout, activation, normF, normA, norm_first, attention2F, device=None, dtype=None) -> None:
        self.factory = {'device': device, 'dtype': dtype}
        super().__init__()
        self.num_heads=nhead

        self.self_attn = attention2F[0](d_model, self.num_heads, dropout=dropout, **attention2F[1], **self.factory)
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward, **self.factory)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model, **self.factory)

        self.norm_first = norm_first
        self.norm1 = normF(d_model, **normA, **self.factory)
        self.norm2 = normF(d_model, **normA, **self.factory)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        # Legacy string support for activation function.
        if isinstance(activation, str):
            activation = _get_activation_fn(activation)

        # We can't test self.activation in forward() in TorchScript,
        # so stash some information about it instead.
        if activation is F.relu or isinstance(activation, tt.nn.ReLU):
            self.activation_relu_or_gelu = 1
        elif activation is F.gelu or isinstance(activation, tt.nn.GELU):
            self.activation_relu_or_gelu = 2
        else:
            self.activation_relu_or_gelu = 0
        self.activation = activation

    def __setstate__(self, state):
        super().__setstate__(state)
        if not hasattr(self, 'activation'):
            self.activation = F.relu


    def forward(
            self,
            src: Tensor,
            src_mask: Optional[Tensor] = None,
            src_key_padding_mask: Optional[Tensor] = None,
            is_causal: bool = False, store_attention=False) -> Tensor:
        r"""Pass the input through the encoder layer.

        Args:
            src: the sequence to the encoder layer (required).
            src_mask: the mask for the src sequence (optional).
            is_causal: If specified, applies a causal mask as src_mask.
              Default: ``False``.
            src_key_padding_mask: the mask for the src keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """
        src_key_padding_mask = F._canonical_mask(
            mask=src_key_padding_mask,
            mask_name="src_key_padding_mask",
            other_type=F._none_or_dtype(src_mask),
            other_name="src_mask",
            target_type=src.dtype
        )

        src_mask = F._canonical_mask(
            mask=src_mask,
            mask_name="src_mask",
            other_type=None,
            other_name="",
            target_type=src.dtype,
            check_other=False,
        )

        # see Fig. 1 of https://arxiv.org/pdf/2002.04745v1.pdf
        why_not_sparsity_fast_path = ''
        if hasattr(self.self_attn, 'learnable'):
            why_not_sparsity_fast_path = f"Learnable Attention."
        elif not src.dim() == 3:
            why_not_sparsity_fast_path = f"input not batched; expected src.dim() of 3 but got {src.dim()}"
        elif self.training:
            why_not_sparsity_fast_path = "training is enabled"
        elif not self.self_attn.batch_first :
            why_not_sparsity_fast_path = "self_attn.batch_first was not True"
        elif not self.self_attn._qkv_same_embed_dim :
            why_not_sparsity_fast_path = "self_attn._qkv_same_embed_dim was not True"
        elif not self.activation_relu_or_gelu:
            why_not_sparsity_fast_path = "activation_relu_or_gelu was not True"
        elif not (self.norm1.eps == self.norm2.eps):
            why_not_sparsity_fast_path = "norm1.eps is not equal to norm2.eps"
        elif src.is_nested and (src_key_padding_mask is not None or src_mask is not None):
            why_not_sparsity_fast_path = "neither src_key_padding_mask nor src_mask are not supported with NestedTensor input"
        elif self.self_attn.num_heads % 2 == 1:
            why_not_sparsity_fast_path = "num_head is odd"
        elif tt.is_autocast_enabled():
            why_not_sparsity_fast_path = "autocast is enabled"
        elif store_attention:
            why_not_sparsity_fast_path="Need to store attention weights."
        if not why_not_sparsity_fast_path:
            tensor_args = (
                src,
                self.self_attn.in_proj_weight,
                self.self_attn.in_proj_bias,
                self.self_attn.out_proj.weight,
                self.self_attn.out_proj.bias,
                self.norm1.weight,
                self.norm1.bias,
                self.norm2.weight,
                self.norm2.bias,
                self.linear1.weight,
                self.linear1.bias,
                self.linear2.weight,
                self.linear2.bias,
            )

            # We have to use list comprehensions below because TorchScript does not support
            # generator expressions.
            if tt.overrides.has_torch_function(tensor_args):
                why_not_sparsity_fast_path = "some Tensor argument has_torch_function"
            elif not all((x.is_cuda or 'cpu' in str(x.device)) for x in tensor_args):
                why_not_sparsity_fast_path = "some Tensor argument is neither CUDA nor CPU"
            elif tt.is_grad_enabled() and any(x.requires_grad for x in tensor_args):
                why_not_sparsity_fast_path = ("grad is enabled and at least one of query or the "
                                              "input/output projection weights or biases requires_grad")

            if not why_not_sparsity_fast_path:
                merged_mask, mask_type = self.self_attn.merge_masks(src_mask, src_key_padding_mask, src)
                return tt._transformer_encoder_layer_fwd(
                    src,
                    self.self_attn.embed_dim,
                    self.self_attn.num_heads,
                    self.self_attn.in_proj_weight,
                    self.self_attn.in_proj_bias,
                    self.self_attn.out_proj.weight,
                    self.self_attn.out_proj.bias,
                    self.activation_relu_or_gelu == 2,
                    self.norm_first,
                    self.norm1.eps,
                    self.norm1.weight,
                    self.norm1.bias,
                    self.norm2.weight,
                    self.norm2.bias,
                    self.linear1.weight,
                    self.linear1.bias,
                    self.linear2.weight,
                    self.linear2.bias,
                    merged_mask,
                    mask_type,
                )


        x = src
        if self.norm_first:
            x = x + self._sa_block(self.norm1(x), src_mask, src_key_padding_mask, is_causal=is_causal, store_attention=store_attention)
            x = x + self._ff_block(self.norm2(x))
        else:
            x = self.norm1(x + self._sa_block(x, src_mask, src_key_padding_mask, is_causal=is_causal, store_attention=store_attention))
            x = self.norm2(x + self._ff_block(x))

        return x

    # self-attention block
    def _sa_block(self, x: Tensor,
                  attn_mask: Optional[Tensor], key_padding_mask: Optional[Tensor], is_causal: bool = False, store_attention=False) -> Tensor:
        x, w = self.self_attn.forward(x, x, x,
                           attn_mask=attn_mask,
                           key_padding_mask=key_padding_mask,
                           is_causal=is_causal,
                           need_weights=store_attention, average_attn_weights=False, )
        if store_attention: self.attention_weights = w
        return self.dropout1(x)

    # feed forward block
    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout2(x)

class TransformerEncoder(nn.Module):
    __constants__ = ['norm']
    
    def __init__(self, coderA, num_layers, norm=None, enable_nested_tensor=True, mask_check=True, dtype=None, device=None):
        self.factory = {'device': device, 'dtype': dtype}
        super().__init__()
        self.num_layers = num_layers
        self.coderA=coderA
        self.layers = nn.ModuleList(  [  TransformerEncoderLayer(**coderA, **self.factory)  for _ in range(self.num_layers) ] ) #_get_clones(encoder_layer, num_layers)
        self.norm = norm
        self.enable_nested_tensor = enable_nested_tensor
        self.mask_check = mask_check
        

    def forward(
            self,
            src: Tensor,
            mask: Optional[Tensor] = None,
            src_key_padding_mask: Optional[Tensor] = None,
            is_causal: Optional[bool] = None, store_attention=False) -> Tensor:

        src_key_padding_mask = F._canonical_mask(
            mask=src_key_padding_mask,
            mask_name="src_key_padding_mask",
            other_type=F._none_or_dtype(mask),
            other_name="mask",
            target_type=src.dtype
        )

        mask = F._canonical_mask(
            mask=mask,
            mask_name="mask",
            other_type=None,
            other_name="",
            target_type=src.dtype,
            check_other=False,
        )

        output = src
        convert_to_nested = False
        first_layer = self.layers[0]
        src_key_padding_mask_for_layers = src_key_padding_mask
        why_not_sparsity_fast_path = ''
        str_first_layer = "self.layers[0]"
        if hasattr(first_layer.self_attn, 'learnable'):
            why_not_sparsity_fast_path = f"{str_first_layer} is Learnable Attention."
        elif not isinstance(first_layer, TransformerEncoderLayer):
            why_not_sparsity_fast_path = f"{str_first_layer} was not TransformerEncoderLayer"
        elif first_layer.norm_first :
            why_not_sparsity_fast_path = f"{str_first_layer}.norm_first was True"
        elif first_layer.training:
            why_not_sparsity_fast_path = f"{str_first_layer} was in training mode"
        elif not first_layer.self_attn.batch_first:
            why_not_sparsity_fast_path = f" {str_first_layer}.self_attn.batch_first was not True"
        elif not first_layer.self_attn._qkv_same_embed_dim:
            why_not_sparsity_fast_path = f"{str_first_layer}.self_attn._qkv_same_embed_dim was not True"
        elif not first_layer.activation_relu_or_gelu:
            why_not_sparsity_fast_path = f" {str_first_layer}.activation_relu_or_gelu was not True"
        elif not (first_layer.norm1.eps == first_layer.norm2.eps) :
            why_not_sparsity_fast_path = f"{str_first_layer}.norm1.eps was not equal to {str_first_layer}.norm2.eps"
        elif not src.dim() == 3:
            why_not_sparsity_fast_path = f"input not batched; expected src.dim() of 3 but got {src.dim()}"
        elif not self.enable_nested_tensor:
            why_not_sparsity_fast_path = "enable_nested_tensor was not True"
        elif src_key_padding_mask is None:
            why_not_sparsity_fast_path = "src_key_padding_mask was None"
        elif (((not hasattr(self, "mask_check")) or self.mask_check)
                and not tt._nested_tensor_from_mask_left_aligned(src, src_key_padding_mask.logical_not())):
            why_not_sparsity_fast_path = "mask_check enabled, and src and src_key_padding_mask was not left aligned"
        elif output.is_nested:
            why_not_sparsity_fast_path = "NestedTensor input is not supported"
        elif mask is not None:
            why_not_sparsity_fast_path = "src_key_padding_mask and mask were both supplied"
        elif first_layer.self_attn.num_heads % 2 == 1:
            why_not_sparsity_fast_path = "num_head is odd"
        elif tt.is_autocast_enabled():
            why_not_sparsity_fast_path = "autocast is enabled"
        elif store_attention:
            why_not_sparsity_fast_path="Need to store attention weights."

        if not why_not_sparsity_fast_path:
            tensor_args = (
                src,
                first_layer.self_attn.in_proj_weight,
                first_layer.self_attn.in_proj_bias,
                first_layer.self_attn.out_proj.weight,
                first_layer.self_attn.out_proj.bias,
                first_layer.norm1.weight,
                first_layer.norm1.bias,
                first_layer.norm2.weight,
                first_layer.norm2.bias,
                first_layer.linear1.weight,
                first_layer.linear1.bias,
                first_layer.linear2.weight,
                first_layer.linear2.bias,
            )

            if tt.overrides.has_torch_function(tensor_args):
                why_not_sparsity_fast_path = "some Tensor argument has_torch_function"
            elif not (src.is_cuda or 'cpu' in str(src.device)):
                why_not_sparsity_fast_path = "src is neither CUDA nor CPU"
            elif tt.is_grad_enabled() and any(x.requires_grad for x in tensor_args):
                why_not_sparsity_fast_path = ("grad is enabled and at least one of query or the "
                                              "input/output projection weights or biases requires_grad")

            if (not why_not_sparsity_fast_path) and (src_key_padding_mask is not None):
                convert_to_nested = True
                output = tt._nested_tensor_from_mask(output, src_key_padding_mask.logical_not(), mask_check=False)
                src_key_padding_mask_for_layers = None

        # Prevent type refinement
        make_causal = (is_causal is True)

        if is_causal is None:
            if mask is not None:
                sz = mask.size(0)
                causal_comparison = tt.triu(
                    tt.ones(sz, sz, device=mask.device) * float('-inf'), diagonal=1
                ).to(mask.dtype)

                if tt.equal(mask, causal_comparison):
                    make_causal = True

        is_causal = make_causal

        for mod in self.layers:
            output = mod(output, src_mask=mask, is_causal=is_causal, src_key_padding_mask=src_key_padding_mask_for_layers, store_attention=store_attention)

        if convert_to_nested:
            output = output.to_padded_tensor(0.)

        if self.norm is not None:
            output = self.norm(output)

        return output


class TransformerDecoderLayer(nn.Module):
    __constants__ = ['batch_first', 'norm_first']

    def __init__(self, d_model, nhead, dim_feedforward, dropout, activation, normF, normA, norm_first, attention2F, device=None, dtype=None) -> None:
        self.factory = {'device': device, 'dtype': dtype}
        super().__init__()
        self.num_heads=nhead

        self.self_attn = attention2F[0](d_model, self.num_heads, dropout=dropout, **attention2F[1], **self.factory)
        self.multihead_attn = attention2F[0](d_model, self.num_heads, dropout=dropout, **attention2F[1], **self.factory)
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward, **self.factory)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model, **self.factory)

        self.norm_first = norm_first
        self.norm1 = normF(d_model, **normA, **self.factory)
        self.norm2 = normF(d_model, **normA, **self.factory)
        self.norm3 = normF(d_model, **normA, **self.factory)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.dropout3 = nn.Dropout(dropout)

        # Legacy string support for activation function.
        if isinstance(activation, str):
            self.activation = _get_activation_fn(activation)
        else:
            self.activation = activation

    def __setstate__(self, state):
        if 'activation' not in state:
            state['activation'] = F.relu
        super().__setstate__(state)

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        tgt_key_padding_mask: Optional[Tensor] = None,
        memory_key_padding_mask: Optional[Tensor] = None,
        tgt_is_causal: bool = False,
        memory_is_causal: bool = False,
        store_attention=False,
    ) -> Tensor:

        x = tgt
        if self.norm_first:
            x = x + self._sa_block(self.norm1(x), tgt_mask, tgt_key_padding_mask, tgt_is_causal,  store_attention=store_attention)
            x = x + self._mha_block(self.norm2(x), memory, memory_mask, memory_key_padding_mask, memory_is_causal,  store_attention=store_attention)
            x = x + self._ff_block(self.norm3(x))
        else:
            x = self.norm1(x + self._sa_block(x, tgt_mask, tgt_key_padding_mask, tgt_is_causal,  store_attention=store_attention))
            x = self.norm2(x + self._mha_block(x, memory, memory_mask, memory_key_padding_mask, memory_is_causal,  store_attention=store_attention))
            x = self.norm3(x + self._ff_block(x))

        return x

    # self-attention block
    def _sa_block(self, x: Tensor,
                  attn_mask: Optional[Tensor], key_padding_mask: Optional[Tensor], is_causal: bool = False,  store_attention=False) -> Tensor:
        x, w = self.self_attn(x, x, x,
                           attn_mask=attn_mask,
                           key_padding_mask=key_padding_mask,
                           is_causal=is_causal,
                           need_weights=store_attention, average_attn_weights=False,)
        if store_attention: self.attention_weights = w
        return self.dropout1(x)

    # multihead attention block
    def _mha_block(self, x: Tensor, mem: Tensor,
                   attn_mask: Optional[Tensor], key_padding_mask: Optional[Tensor], is_causal: bool = False,  store_attention=False) -> Tensor:
        x, w = self.multihead_attn(x, mem, mem,
                                attn_mask=attn_mask,
                                key_padding_mask=key_padding_mask,
                                is_causal=is_causal,
                                need_weights=store_attention, average_attn_weights=False,)
        if store_attention: self.attention_weights_cross = w
        return self.dropout2(x)

    # feed forward block
    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout3(x)

class TransformerDecoder(nn.Module):

    __constants__ = ['norm']

    def __init__(self, coderA, num_layers, norm=None, dtype=None, device=None):
        self.factory = {'device': device, 'dtype': dtype}
        super().__init__()
        self.num_layers = num_layers
        self.coderA=coderA
        self.layers = nn.ModuleList(  [  TransformerDecoderLayer(**coderA, **self.factory)  for _ in range(self.num_layers) ] )
        self.norm = norm
        

    def forward(self, tgt: Tensor, memory: Tensor, multi_memory: bool=False, tgt_mask: Optional[Tensor] = None,
                memory_mask: Optional[Tensor] = None, tgt_key_padding_mask: Optional[Tensor] = None,
                memory_key_padding_mask: Optional[Tensor] = None, store_attention=False) -> Tensor:

        output = tgt

        if multi_memory:
            # expects list of  memory, memory_mask, memory_key_padding_mask
            for mod, mem, mem_mask, mem_key_pad_mask in zip(self.layers, memory, memory_mask, memory_key_padding_mask):
                output = mod(output, mem, tgt_mask=tgt_mask,
                            memory_mask=mem_mask,
                            tgt_key_padding_mask=tgt_key_padding_mask,
                            memory_key_padding_mask=mem_key_pad_mask,
                            store_attention=store_attention)
        else:
            for mod in self.layers:
                output = mod(output, memory, tgt_mask=tgt_mask,
                            memory_mask=memory_mask,
                            tgt_key_padding_mask=tgt_key_padding_mask,
                            memory_key_padding_mask=memory_key_padding_mask,
                            store_attention=store_attention)

        if self.norm is not None:
            output = self.norm(output)

        return output


class TransformerCrossLayer(nn.Module):
    __constants__ = ['batch_first', 'norm_first']

    def __init__(self, d_model, nhead, dim_feedforward, dropout, activation, normF, normA, norm_first, attention2F, device=None, dtype=None) -> None:
        self.factory = {'device': device, 'dtype': dtype}
        super().__init__()
        self.multihead_attn = attention2F[0](d_model, self.num_heads, dropout=dropout, **attention2F[1], **self.factory)
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward, **self.factory)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model, **self.factory)

        self.norm_first = norm_first
        #self.norm1 = LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)
        self.norm2 = normF(d_model, **normA, **self.factory)
        self.norm3 = normF(d_model, **normA, **self.factory)
        #self.dropout1 = Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.dropout3 = nn.Dropout(dropout)

        # Legacy string support for activation function.
        if isinstance(activation, str):
            self.activation = _get_activation_fn(activation)
        else:
            self.activation = activation

    def __setstate__(self, state):
        if 'activation' not in state:
            state['activation'] = F.relu
        super().__setstate__(state)

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        #tgt_mask: Optional[Tensor] = None,
        memory_mask: Optional[Tensor] = None,
        #tgt_key_padding_mask: Optional[Tensor] = None,
        memory_key_padding_mask: Optional[Tensor] = None,
        #tgt_is_causal: bool = False,
        memory_is_causal: bool = False,
        store_attention=False,
    ) -> Tensor:

        x = tgt
        if self.norm_first:
            #x = x + self._sa_block(self.norm1(x), tgt_mask, tgt_key_padding_mask, tgt_is_causal,  store_attention=store_attention)
            x = x + self._mha_block(self.norm2(x), memory, memory_mask, memory_key_padding_mask, memory_is_causal,  store_attention=store_attention)
            x = x + self._ff_block(self.norm3(x))
        else:
            #x = self.norm1(x + self._sa_block(x, tgt_mask, tgt_key_padding_mask, tgt_is_causal,  store_attention=store_attention))
            x = self.norm2(x + self._mha_block(x, memory, memory_mask, memory_key_padding_mask, memory_is_causal,  store_attention=store_attention))
            x = self.norm3(x + self._ff_block(x))

        return x


    # multihead attention block
    def _mha_block(self, x: Tensor, mem: Tensor,
                   attn_mask: Optional[Tensor], key_padding_mask: Optional[Tensor], is_causal: bool = False,  store_attention=False) -> Tensor:
        x, w = self.multihead_attn(x, mem, mem,
                                attn_mask=attn_mask,
                                key_padding_mask=key_padding_mask,
                                is_causal=is_causal,
                                need_weights=store_attention, average_attn_weights=False,)
        if store_attention: self.attention_weights_cross = w
        return self.dropout2(x)

    # feed forward block
    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout3(x)

class TransformerCross(nn.Module):

    __constants__ = ['norm']

    def __init__(self, coderA, num_layers, norm=None, dtype=None, device=None):
        self.factory = {'device': device, 'dtype': dtype}
        super().__init__()
        self.num_layers = num_layers
        self.coderA = coderA
        self.layers = nn.ModuleList(  [  TransformerCrossLayer(**coderA, **self.factory)  for _ in range(self.num_layers) ] )
        self.norm = norm


    def forward(self, tgt: Tensor, memory: Tensor, multi_memory: bool=False, tgt_mask: Optional[Tensor] = None,
                memory_mask: Optional[Tensor] = None, tgt_key_padding_mask: Optional[Tensor] = None,
                memory_key_padding_mask: Optional[Tensor] = None, store_attention=False) -> Tensor:

        output = tgt

        if multi_memory:
            for mod,mem,mem_mask,mem_key_pad_mask in zip(self.layers, memory, memory_mask, memory_key_padding_mask):
                output = mod(output, mem, tgt_mask=tgt_mask,
                            memory_mask=mem_mask,
                            tgt_key_padding_mask=tgt_key_padding_mask,
                            memory_key_padding_mask=mem_key_pad_mask,
                            store_attention=store_attention)
        else:
            for mod in self.layers:
                output = mod(output, memory, tgt_mask=tgt_mask,
                            memory_mask=memory_mask,
                            tgt_key_padding_mask=tgt_key_padding_mask,
                            memory_key_padding_mask=memory_key_padding_mask,
                            store_attention=store_attention)

        if self.norm is not None:
            output = self.norm(output)

        return output


class Encoder(TransformerEncoder):

    def __init__(self, vocab_size, pose,
                 coderA, num_layers, norm=None, 
                 #enable_nested_tensor=True, 
                 #mask_check=True, 
                 dtype=None, device=None):
        super().__init__(coderA, num_layers, norm, True, True, dtype, device)
        #self.coderA['vocab_size'] = vocab_size
        self.embeder = nn.Embedding(vocab_size, self.coderA['d_model'], **self.factory)
        self.pose=pose
        self.do_store_attention(False)


    def do_store_attention(self, do_store): self.store_attention=do_store

    def forward(self, 
                src: Tensor, 
                mask: Tensor | None = None, 
                src_key_padding_mask: Tensor | None = None, 
                #is_causal: bool | None = None, 
                #store_attention=False
                ) -> Tensor:
        srcE = self.pose(self.embeder(src))
        return super().forward(srcE, mask, src_key_padding_mask, None, self.store_attention)


    # @tt.no_grad()
    # def view_attention(self, batch_index=None, head_index=None, width=16, values=False, ticks=False, verbose=0, save='', title='', **matshow):
        
    #     #plt.matshow(cdec.decoder.layers[0].attention_weights_cross[0,:,:])
    #     fig, axs = plt.subplots(self.num_layers, 1, figsize=(width, self.num_layers*width))
    #     if not title: title=f'{__class__}'
    #     fig.suptitle(f'{title}')
    #     single = self.num_layers==1


    #     for l in range(self.num_layers):
    #         # batch_size, n_heads, SEQ, SEQ
    #         if batch_index is None:
    #             if head_index is None:
    #                 w = self.layers[l].attention_weights.detach().cpu().mean(dim=0)
    #             else:
    #                 w = self.layers[l].attention_weights[head_index].detach().cpu()
    #         else:
    #             if head_index is None:
    #                 w = self.layers[l].attention_weights[batch_index, :].detach().cpu().mean(dim=0)
    #             else:
    #                 w = self.layers[l].attention_weights[batch_index, head_index].detach().cpu()

    #         ax = axs if single else axs[l]
    #         ax.matshow(w.numpy(), **matshow)
    #         ax.set_xlabel(f'layer: {l+1}')
    #         if ticks:
    #             ax.set_xticks(range(w.shape[1]))
    #             ax.set_yticks(range(w.shape[0]))
    #         if values:
    #             for i in range(w.shape[0]):
    #                 for j in range(w.shape[1]):
    #                     ax.text(j, i, '{:0.2f}'.format(w[i,j].item()), ha='center', va='center',
    #                         bbox=dict(boxstyle='round', facecolor='white', edgecolor='0.3'))
    #         if verbose: print(f'Layer: {l} :: \n{w}')

    #     if save:
    #         fig.savefig(save)
    #         plt.close()
    #     else:
    #         plt.show()

class Decoder(TransformerDecoder):
    def __init__(self, vocab_size, pose,
                 coderA, num_layers, norm=None, 
                 dtype=None, device=None):
        super().__init__(coderA, num_layers, norm, dtype, device)
        #self.coderA['vocab_size'] = vocab_size
        self.embeder = nn.Embedding(vocab_size, self.coderA['d_model'], **self.factory)
        self.pose=pose
        self.do_store_attention(False)
    
    def do_store_attention(self, do_store): self.store_attention=do_store

    def forward(self, 
                tgt: Tensor, 
                memory: Tensor, 
                multi_memory: bool = False, 
                tgt_mask: Tensor | None = None, 
                memory_mask: Tensor | None = None, 
                tgt_key_padding_mask: Tensor | None = None, 
                memory_key_padding_mask: Tensor | None = None, 
                #store_attention=False
                ) -> Tensor:
        tgtE = self.pose(self.embeder(tgt))
        return super().forward(tgtE, memory, multi_memory, tgt_mask, memory_mask, tgt_key_padding_mask, memory_key_padding_mask, self.store_attention)
    


    # @tt.no_grad()
    # def view_attention(self, cross,  batch_index=None, head_index=None, width=16, values=False, ticks=False, verbose=0, save='', title='', **matshow):
        
    #     #plt.matshow(cdec.decoder.layers[0].attention_weights_cross[0,:,:])
    #     fig, axs = plt.subplots(self.num_layers, 1, figsize=(width, self.num_layers*width))
    #     if not title: title=f'{__class__}'
    #     fig.suptitle(f'{title}')
    #     single = self.num_layers==1


    #     for l in range(self.num_layers):
    #         # batch_size, n_heads, SEQ, SEQ
    #         if batch_index is None:
    #             if head_index is None:
    #                 if cross:
    #                     w = self.layers[l].attention_weights_cross.detach().cpu().mean(dim=0)
    #                 else:
    #                     w = self.layers[l].attention_weights.detach().cpu().mean(dim=0)
    #             else:
    #                 if cross:
    #                     w = self.layers[l].attention_weights_cross[head_index].detach().cpu()
    #                 else:
    #                     w = self.layers[l].attention_weights[head_index].detach().cpu()
    #         else:
    #             if head_index is None:
    #                 if cross:
    #                     w = self.layers[l].attention_weights_cross[batch_index, :].detach().cpu().mean(dim=0)
    #                 else:
    #                     w = self.layers[l].attention_weights[batch_index, :].detach().cpu().mean(dim=0)
    #             else:
    #                 if cross:
    #                     w = self.layers[l].attention_weights_cross[batch_index, head_index].detach().cpu()
    #                 else:
    #                     w = self.layers[l].attention_weights[batch_index, head_index].detach().cpu()

    #         ax = axs if single else axs[l]
    #         ax.matshow(w.numpy(), **matshow)
    #         ax.set_xlabel(f'layer: {l+1}')
    #         if ticks:
    #             ax.set_xticks(range(w.shape[1]))
    #             ax.set_yticks(range(w.shape[0]))
    #         if values:
    #             for i in range(w.shape[0]):
    #                 for j in range(w.shape[1]):
    #                     ax.text(j, i, '{:0.2f}'.format(w[i,j].item()), ha='center', va='center',
    #                         bbox=dict(boxstyle='round', facecolor='white', edgecolor='0.3'))
    #         if verbose: print(f'Layer: {l} :: \n{w}')

    #     if save:
    #         fig.savefig(save)
    #         plt.close()
    #     else:
    #         plt.show()





import torch as tt
import torch.nn as nn
from torch.nn.init import xavier_uniform_
import matplotlib.pyplot as plt

__all__ = [ 'Mod', 'Aview',]


class Mod:
    
    @staticmethod
    def Count(module, requires_grad=None): 
        r""" Counts the total number of parameters (numel) in a params
        
        :param requires_grad: 
            if None, counts all parameters
            if True, counts trainable parameters
            if False, counts non-trainiable (frozen) parameters
        """
        return sum( ([ p.numel() for p in module.parameters() ]) if requires_grad is None else \
                    ([ p.numel() for p in module.parameters()    if p.requires_grad is requires_grad ]) )

    @staticmethod
    def Show(module, values:bool=False):
        r""" Prints the parameters of a params
        
        :param values: if True, prints the full tensors otherwise prints only shape
        """
        nos_trainable, nos_frozen = 0, 0
        print('=====================================')
        for i,p in enumerate(module.parameters()):
            iparam = p.numel()
            if p.requires_grad:
                nos_trainable += iparam
            else:
                nos_frozen += iparam
            print(f'#[{i}]\tShape[{p.shape}]\tParams: {iparam}\tTrainable: {p.requires_grad}')
            if values: 
                print('=====================================')
                print(f'{p}')
                print('=====================================')
        print(f'\nTotal Parameters: {nos_trainable+nos_frozen}\tTrainable: {nos_trainable}\tFrozen: {nos_frozen}')
        print('=====================================')
        return 

    @staticmethod
    def State(module, values=False):
        r""" prints the parameters using `nn.Module.parameters` iterator, use `values=True` to print full parameter tensor """
        sd = module.state_dict()
        for i,(k,v) in enumerate(sd.items()):
            print(f'#[{i+1}]\t[{k}]\tShape[{v.shape}]')
            if values: print(f'{v}')
        return 

    @staticmethod
    def Dense(in_dim, layer_dims, out_dim, 
            actFs, bias=True, dtype=None, device=None ):
        r"""
        Creats a stack of fully connected (dense) layers which is usually connected at end of other networks
        Args:
            in_dim          `integer`       : in_features or input_size
            layer_dims      `List/Tuple`    : size of hidden layers
            out_dim         `integer`       : out_features or output_size
            actFs           `nn.Module`     : activation function at hidden layer
            bias            `bool`          : if True, uses bias at hidden layers

        Returns:
            `nn.Module` : an instance of nn.Sequential
        """
        layers = []
        # first layer
        layers.append(nn.Linear(in_dim, layer_dims[0], bias=bias, dtype=dtype, device=device))
        if actFs: layers.append(actFs.pop(0))
        # remaining layers
        for i in range(len(layer_dims)-1):
            layers.append(nn.Linear(layer_dims[i], layer_dims[i+1], bias=bias, dtype=dtype, device=device))
            if actFs: layers.append(actFs.pop(0))
        # last layer
        layers.append(nn.Linear(layer_dims[-1], out_dim, bias=bias, dtype=dtype, device=device))
        if actFs: layers.append(actFs.pop(0))
        return nn.Sequential( *layers )

    @staticmethod
    def Clones(module, n_copies:int):
        from io import BytesIO
        r""" Replicates a params by storing it in a buffer and retriving many copies
        NOTE: this will preserve the ```require_grad``` attribute on all tensors. """
        #from io import BytesIO
        if n_copies<1: return None
        buffer = BytesIO()
        tt.save(module, buffer)
        model_copies = []
        for _ in range(n_copies):
            buffer.seek(0)
            model_copy = tt.load(buffer)
            model_copies.append(model_copy)
        buffer.close()
        del buffer
        return model_copies

    @staticmethod
    def Clone(module): return __class__.Clones(module, 1).pop()

    @staticmethod
    def SetGrad(module, requires:bool, *names):
        r""" Sets requires_grad attribute on tensors in params
        if no names are provided, sets requires_grad on all tensors 
        NOTE: careful with *names, if a buffer's name is provided
            and it is in the state_dict then its grad will be enabled
            which is undesirable.
            not providing any names will target the parameters only
        """
        if names: # if we know which params to freeze, we can provide them
            state_dict = module.state_dict() 
            for n in names: state_dict[n].requires_grad_(requires)
        else: # if we want to do on all params
            for p in module.parameters(): p.requires_grad_(requires)
        return module

    @staticmethod
    def CausalMask(block_size, diagonal=1, dtype=None, device=None): 
        return tt.triu(tt.full((block_size, block_size), -tt.inf, dtype=dtype, device=device), diagonal=diagonal)

    @staticmethod
    def CausalCrossMask(d_block_size, e_block_size, diagonal=1, dtype=None, device=None): 
        return tt.triu(tt.full((d_block_size, e_block_size), -tt.inf, dtype=dtype, device=device), diagonal=diagonal)

    # "Secrete Functions"
    @staticmethod
    @tt.no_grad()
    def reset_parameters_xavier_uniform(module):
        for p in module.parameters():
            if p.dim() > 1: xavier_uniform_(p)

class Aview:
    dca = dict(width=4, values=False, ticks=True, verbose=0,  cmap='hot',      vmin=0, vmax=1)
    dsa = dict(width=4, values=False, ticks=True, verbose=0,  cmap='binary',   vmin=0, vmax=1)
    esa = dict(width=4, values=False, ticks=True, verbose=0,  cmap='copper',   vmin=0, vmax=1)


    @tt.no_grad()
    def view_attention(coder, cross,  batch_index=None, head_index=None, width=16, values=False, ticks=False, verbose=0, save='', title='', **matshow):
        
        #plt.matshow(cdec.decoder.layers[0].attention_weights_cross[0,:,:])
        fig, axs = plt.subplots(coder.num_layers, 1, figsize=(width, coder.num_layers*width))
        if title: fig.suptitle(f'{title}')
        single = coder.num_layers==1


        for l in range(coder.num_layers):
            # batch_size, n_heads, SEQ, SEQ
            if batch_index is None:
                if head_index is None:
                    if cross:
                        w = coder.layers[l].attention_weights_cross.detach().cpu().mean(dim=0)
                    else:
                        w = coder.layers[l].attention_weights.detach().cpu().mean(dim=0)
                else:
                    if cross:
                        w = coder.layers[l].attention_weights_cross[head_index].detach().cpu()
                    else:
                        w = coder.layers[l].attention_weights[head_index].detach().cpu()
            else:
                if head_index is None:
                    if cross:
                        w = coder.layers[l].attention_weights_cross[batch_index, :].detach().cpu().mean(dim=0)
                    else:
                        w = coder.layers[l].attention_weights[batch_index, :].detach().cpu().mean(dim=0)
                else:
                    if cross:
                        w = coder.layers[l].attention_weights_cross[batch_index, head_index].detach().cpu()
                    else:
                        w = coder.layers[l].attention_weights[batch_index, head_index].detach().cpu()

            ax = axs if single else axs[l]
            ax.matshow(w.numpy(), **matshow)
            ax.set_xlabel(f'layer: {l+1}')
            if ticks:
                ax.set_xticks(range(w.shape[1]))
                ax.set_yticks(range(w.shape[0]))
            if values:
                for i in range(w.shape[0]):
                    for j in range(w.shape[1]):
                        ax.text(j, i, '{:0.2f}'.format(w[i,j].item()), ha='center', va='center',
                            bbox=dict(boxstyle='round', facecolor='white', edgecolor='0.3'))
            if verbose: print(f'Layer: {l} :: \n{w}')

        if save:
            fig.savefig(save)
            plt.close()
        else:
            plt.show()

    def view_attention_encoder(former, batch_index=None, head_index=None, width=16, values=False, ticks=False, verbose=0, save='', title='', **matshow):
        __class__.view_attention(former.encoder, False, batch_index=batch_index, head_index=head_index, width=width, 
           values=values, ticks=ticks, verbose=verbose, save=save, title=title, **matshow)    

    def view_attention_decoder(former, cross, batch_index=None, head_index=None, width=16, values=False, ticks=False, verbose=0, save='', title='', **matshow):
        __class__.view_attention(former.decoder, cross, batch_index=batch_index, head_index=head_index, width=width, 
                values=values, ticks=ticks, verbose=verbose, save=save, title=title, **matshow) 

    def view_attention_encoders(former, batch_index=None, head_index=None, width=16, values=False, ticks=False, verbose=0, save='', title='', **matshow):
        for encoder in former.encoders: __class__.view_attention(encoder, False, batch_index=batch_index, head_index=head_index, width=width, 
           values=values, ticks=ticks, verbose=verbose, save=save, title=title, **matshow)    


    @tt.no_grad()
    def view_mh_attention(coder, cross,  batch_index=None, width=16, values=False, ticks=False, verbose=0, save='', title='', **matshow):

        for l in range(coder.num_layers):
            num_heads = coder.layers[l].num_heads
            fig, axs = plt.subplots(1,  num_heads, figsize=(width*num_heads, width))
            if title:  fig.suptitle(f'{title}')
            single_head = num_heads==1
            # batch_size, n_heads, SEQ, SEQ
            for h in range(num_heads):
                if batch_index is None:
                    if cross:
                        w = coder.layers[l].attention_weights_cross[h].detach().cpu()
                    else:
                        w = coder.layers[l].attention_weights[h].detach().cpu()
                else:
                    if cross:
                        w = coder.layers[l].attention_weights_cross[batch_index, h].detach().cpu()
                    else:
                        w = coder.layers[l].attention_weights[batch_index, h].detach().cpu()

                if single_head: ax = axs
                else: ax = axs[h]

                
                ax.matshow(w.numpy(), **matshow)
                ax.set_xlabel(f'layer: {l+1}, head: {h+1}')
                if ticks:
                    ax.set_xticks(range(w.shape[1]))
                    ax.set_yticks(range(w.shape[0]))
                if values:
                    for i in range(w.shape[0]):
                        for j in range(w.shape[1]):
                            ax.text(j, i, '{:0.2f}'.format(w[i,j].item()), ha='center', va='center',
                                bbox=dict(boxstyle='round', facecolor='white', edgecolor='0.3'))
                if verbose: print(f'layer: {l+1}, head: {h+1} :: \n{w}')

            if save: fig.savefig(f'{save}_{l}.png')
            else: plt.show()
            plt.close()


    def view_mh_attention_encoder(former, batch_index=None, width=16, values=False, ticks=False, verbose=0, save='', title='', **matshow):
        __class__.view_mh_attention(former.encoder, False, batch_index=batch_index,  width=width, 
           values=values, ticks=ticks, verbose=verbose, save=save, title=title, **matshow)    

    def view_mh_attention_decoder(former, cross, batch_index=None,  width=16, values=False, ticks=False, verbose=0, save='', title='', **matshow):
        __class__.view_mh_attention(former.decoder, cross, batch_index=batch_index,  width=width, 
                values=values, ticks=ticks, verbose=verbose, save=save, title=title, **matshow) 

    def view_mh_attention_encoders(former, batch_index=None,  width=16, values=False, ticks=False, verbose=0, save='', title='', **matshow):
        for encoder in former.encoders: __class__.view_mh_attention(encoder, False, batch_index=batch_index, width=width, 
           values=values, ticks=ticks, verbose=verbose, save=save, title=title, **matshow)    


#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=







__all__ = ['Optimizers']

class Optimizers:

    @staticmethod
    def GPTAdamW(model, device, weight_decay=1e-2, learning_rate=1e-3,  betas=(0.9, 0.999), eps=1e-8, verbose=False): 
        import inspect
        from torch.optim import AdamW
        # start with all of the candidate parameters
        param_dict = {pn: p for pn, p in model.named_parameters()}
        # filter out those that do not require grad
        param_dict = {pn: p for pn, p in param_dict.items() if p.requires_grad}
        # create optim groups. Any parameters that is 2D will be weight decayed, otherwise no.
        # i.e. all weight tensors in matmuls + embeddings decay, all biases and layernorms don't.
        decay_params = [p for n, p in param_dict.items() if p.dim() >= 2]
        nodecay_params = [p for n, p in param_dict.items() if p.dim() < 2]
        optim_groups = [
            {'params': decay_params, 'weight_decay': weight_decay},
            {'params': nodecay_params, 'weight_decay': 0.0}
        ]
        num_decay_params = sum(p.numel() for p in decay_params)
        num_nodecay_params = sum(p.numel() for p in nodecay_params)
        if verbose:
            print(f"num decayed parameter tensors: {len(decay_params)}, with {num_decay_params:,} parameters")
            print(f"num non-decayed parameter tensors: {len(nodecay_params)}, with {num_nodecay_params:,} parameters")
        # Create AdamW optimizer and use the fused version if it is available
        fused_available = 'fused' in inspect.signature(AdamW).parameters
        use_fused = fused_available and device == 'cuda'
        extra_args = dict(fused=True) if use_fused else dict()
        optimizer = AdamW(optim_groups, lr=learning_rate, betas=betas, eps=eps, **extra_args)
        if verbose: print(f"using fused AdamW: {use_fused}")

        return optimizer




import torch as tt
import torch.nn as nn
from torch import Tensor

from .basic import Mod
from .torchtf import Encoder, Decoder

__all__ = [ 
    'Seq2SeqTransformer', 
    'MultiSeq2SeqTransformer', 
    'SeqEncodingTransformer', 
    'MultiSeqEncodingTransformer'
]


class Seq2SeqTransformer(nn.Module):  
        
    def do_store_attention(self, do_store, encoder=True, decoder=True):
        if decoder: self.decoder.do_store_attention(do_store)
        if encoder: self.encoder.do_store_attention(do_store)    

    def __init__(self,custom_encoder:Encoder, custom_decoder:Decoder,
                 n_outputs:int,
                 dense_layer_dims:list, 
                 dense_actFs:list, 
                 dense_bias=True, 
                 xavier_init=False,
                 device=None, dtype=None) -> None:
        assert custom_encoder.coderA['d_model']==custom_decoder.coderA['d_model'], f'Expecting same embedding dimension!'
        self.factory = {'device': device, 'dtype': dtype}
        super().__init__()
        self.encoder = custom_encoder
        self.decoder = custom_decoder
        self.embed_size = custom_decoder.coderA['d_model']
        self.n_outputs = n_outputs
        self.dense = Mod.Dense(
            in_dim=self.embed_size, layer_dims=dense_layer_dims, out_dim=self.n_outputs,
            actFs=dense_actFs, bias=dense_bias, **self.factory)
        if xavier_init: Mod.reset_parameters_xavier_uniform(self)

    def forward(self, src: Tensor, tgt: Tensor, src_mask = None, tgt_mask  = None,
                src_key_padding_mask = None,
                tgt_key_padding_mask = None, 
                memory_mask=None,
                memory_key_padding_mask=None,
            ) -> Tensor:

        memory = self.encoder.forward(
            src=src,
            mask=src_mask,
            src_key_padding_mask=src_key_padding_mask)
        
        output = self.decoder.forward(
            tgt=tgt, 
            memory=memory,
            multi_memory=False,
            tgt_mask=tgt_mask,
            memory_mask=memory_mask,
            tgt_key_padding_mask=tgt_key_padding_mask,
            memory_key_padding_mask=memory_key_padding_mask)
        return self.dense(output)

class MultiSeq2SeqTransformer(nn.Module): 
        
    def do_store_attention(self, do_store, encoder=True, decoder=True):
        if decoder: self.decoder.do_store_attention(do_store)
        if encoder: 
            for encoder in self.encoders: encoder.do_store_attention(do_store)    

    def __init__(self,custom_encoders:list[Encoder], custom_decoder:Decoder,
                 catenate:bool,
                 n_outputs:int,
                 dense_layer_dims:list, 
                 dense_actFs:list, 
                 dense_bias=True, 
                 edl_mapping=None,
                 xavier_init=False,
                 device=None, dtype=None) -> None:
        self.catenate = catenate
        if catenate:
            lenc = len(custom_encoders)
            for custom_encoder in custom_encoders:
                assert custom_encoder.coderA['d_model']*lenc==custom_decoder.coderA['d_model'], f'Expecting same embedding dimension!'
        else:
            for custom_encoder in custom_encoders:
                assert custom_encoder.coderA['d_model']==custom_decoder.coderA['d_model'], f'Expecting same embedding dimension!'
        self.factory = {'device': device, 'dtype': dtype}
        super().__init__()
        self.encoders = nn.ModuleList(custom_encoders)
        self.n_encoders=len(self.encoders)
        self.decoder = custom_decoder
        self.embed_size = custom_decoder.coderA['d_model']
        self.n_outputs = n_outputs
        self.set_encoder_decoder_layer_mapping(edl_mapping)
        self.dense = Mod.Dense(
            in_dim=self.embed_size, layer_dims=dense_layer_dims, out_dim=self.n_outputs,
            actFs=dense_actFs, bias=dense_bias, **self.factory)
        self.default_null_mask=[None for _ in range(self.n_encoders)]
        self.default_mem_mask=[None for _ in range(self.decoder.num_layers)]
        
        if xavier_init: Mod.reset_parameters_xavier_uniform(self)


    def set_encoder_decoder_layer_mapping(self, mapping):
        if mapping is None:  self.encoder_decoder_layer_mapping = None
        else:
            assert len(mapping) == self.decoder.num_layers, f'Require one mapping index for each decoder layer : {self.decoder.num_layers}'
            for i,m in enumerate(mapping): assert isinstance(m, int) and m>=0 and m<=self.n_encoders, f'Invalid mapping {m=} and index {i=}'
            self.encoder_decoder_layer_mapping = mapping

    def forward(self, src: Tensor, tgt: Tensor, src_mask = None, tgt_mask  = None,
                src_key_padding_mask = None,
                tgt_key_padding_mask = None, 
                memory_mask=None,
                memory_key_padding_mask=None,
            ) -> Tensor:
        if src_mask is None: src_mask=self.default_null_mask
        if src_key_padding_mask is None: src_key_padding_mask=self.default_null_mask
        if self.catenate:
            memory = tt.cat([encoder.forward(
                        src=isrc,
                        mask=isrc_mask, 
                        src_key_padding_mask=isrc_key_padding_mask) for encoder,isrc,isrc_mask,isrc_key_padding_mask in \
                            zip(self.encoders, src, src_mask, src_key_padding_mask)], dim=-1)
        else:
            memory_ = [encoder.forward(
                        src=isrc,
                        mask=isrc_mask, 
                        src_key_padding_mask=isrc_key_padding_mask) for encoder,isrc,isrc_mask,isrc_key_padding_mask in \
                            zip(self.encoders, src, src_mask,src_key_padding_mask)]
            
            memory =  [ memory_[encoder_index] for encoder_index in self.encoder_decoder_layer_mapping ]  if self.encoder_decoder_layer_mapping else memory_
    
        # NOTE: for catenate=True, the block_size of all encoder must be the same, 
        # but this not the case of catenate=False where encoders can have different sequence lengths
        output = self.decoder.forward(
            tgt=tgt, 
            memory=memory,
            multi_memory=not self.catenate,
            tgt_mask=tgt_mask,
            memory_mask=memory_mask if self.catenate else (self.default_mem_mask if memory_mask is None else memory_mask),
            tgt_key_padding_mask=tgt_key_padding_mask,
            memory_key_padding_mask=memory_key_padding_mask if self.catenate else (self.default_mem_mask if memory_key_padding_mask is None else memory_key_padding_mask),)
        return self.dense(output)

class SeqEncodingTransformer(nn.Module): 

    def do_store_attention(self, do_store): self.encoder.do_store_attention(do_store)    

    def __init__(self,custom_encoder:Encoder,
                 n_outputs:int,
                 dense_layer_dims:list, 
                 dense_actFs:list, 
                 dense_bias=True, 
                 xavier_init=False,
                 device=None, dtype=None) -> None:
        #assert custom_encoder.coderA['d_model']==custom_decoder.coderA['d_model'], f'Expecting same embedding dimension!'
        self.factory = {'device': device, 'dtype': dtype}
        super().__init__()
        self.encoder = custom_encoder
        self.embed_size = custom_encoder.coderA['d_model']
        self.n_outputs = n_outputs
        self.dense = Mod.Dense(
            in_dim=self.embed_size, layer_dims=dense_layer_dims, out_dim=self.n_outputs,
            actFs=dense_actFs, bias=dense_bias, **self.factory)
        if xavier_init: Mod.reset_parameters_xavier_uniform(self)

    def forward(self, src: Tensor, 
                src_mask = None, 
                src_key_padding_mask = None,
            ) -> Tensor:
        return self.dense(self.encoder.forward(
            src=src,
            mask=src_mask,
            src_key_padding_mask=src_key_padding_mask))

class MultiSeqEncodingTransformer(nn.Module):

    def do_store_attention(self, do_store):
        for encoder in self.encoders: encoder.do_store_attention(do_store)    

    def __init__(self,custom_encoders:list[Encoder],
                 catenate:bool,
                 n_outputs:int,
                 dense_layer_dims:list, 
                 dense_actFs:list, 
                 dense_bias=True, 
                 xavier_init=False,
                 device=None, dtype=None) -> None:
        self.catenate = catenate
        for custom_encoder in custom_encoders:
            assert custom_encoder.coderA['d_model']==custom_encoders[0].coderA['d_model'], f'Expecting same embedding dimension!'
        self.factory = {'device': device, 'dtype': dtype}
        super().__init__()
        self.encoders = nn.ModuleList(custom_encoders)
        self.n_encoders=len(self.encoders)
        self.embed_size = custom_encoders[0].coderA['d_model']
        self.n_outputs = n_outputs
        self.dense = Mod.Dense(
            in_dim=self.embed_size*self.n_encoders, layer_dims=dense_layer_dims, out_dim=self.n_outputs,
            actFs=dense_actFs, bias=dense_bias, **self.factory)
        if xavier_init: Mod.reset_parameters_xavier_uniform(self)


    def forward(self, 
                src: Tensor, 
                src_mask = None, 
                src_key_padding_mask = None,
            ) -> Tensor:
        if src_mask is None: src_mask=self.default_null_mask
        if src_key_padding_mask is None: src_key_padding_mask=self.default_null_mask
        output = tt.cat([encoder.forward(
                    src=isrc,
                    mask=isrc_mask, 
                    src_key_padding_mask=isrc_key_padding_mask) for encoder,isrc,isrc_mask,isrc_key_padding_mask in zip(self.encoders, src, src_mask,src_key_padding_mask)], dim=-1)

        return self.dense(output)





import numpy as np
import torch as tt
import torch.nn as nn
from torch.nn import functional as ff
from math import log

__all__ = ['Pose', 'Norm','Vocab',]


class Pose:

    class FixedSinusoidal(nn.Module):

        def __init__(self, input_size: int, block_size: int, dim_constant:float=1e4, dtype=None, device=None):
            super().__init__()
            self.factory = dict(dtype=dtype, device=device) 
            self.input_size, self.block_size = input_size, block_size
            # self.dropout = nn.Dropout(p=dropout)  # we do not need dropout in embedding
            position = tt.arange(block_size).unsqueeze(1)
            div_term = tt.exp(tt.arange(0, input_size, 2) * (-log(dim_constant)/ input_size))
            embedding = tt.zeros((block_size, 1, input_size), dtype=dtype)
            embedding[:, 0, 0::2] = tt.sin(position * div_term)
            embedding[:, 0, 1::2] = tt.cos(position * div_term)
            embedding.swapaxes_(1,0)
            embedding = embedding.to(**self.factory)
            self.register_buffer('embedding', embedding) #<--- optinal

        def forward(self, x): return x + self.embedding
        
    class TrainableLinear(nn.Module):

        def __init__(self, input_size: int, block_size: int, dtype=None, device=None ):
            super().__init__()
            self.factory = dict(dtype=dtype, device=device)
            self.input_size, self.block_size = input_size, block_size
            # self.dropout = nn.Dropout(p=dropout)  # we do not need dropout in embedding
            self.embedding = nn.Embedding(block_size, input_size, **self.factory)
            self.position = tt.arange(0, block_size, 1, dtype=tt.int32, device=device) #<-- no need to unsqueeze, will broadcast
            #NOTE: call self.embedding(self.position) every time because the embedding weights get trained

        def forward(self, x): return x +  self.embedding(self.position)


class Norm:

    class _Norm(nn.Module):

        def __init__(self, embed_dim, bias=True, eps=1e-6, dtype=None, device=None):
            super().__init__()
            self.factory = dict(dtype=dtype, device=device)
            self.embed_dim=embed_dim
            self.eps = eps
            self.weight = nn.Parameter(tt.ones(self.embed_dim, **self.factory))
            self.bias = nn.Parameter(tt.zeros(self.embed_dim, **self.factory)) if bias else None 

    class rmsN(_Norm):

        def __init__(self, embed_dim, bias=True, eps=1e-6, dtype=None, device=None):
            super().__init__(embed_dim, bias, eps, dtype, device)
            if self.bias is None: self.bias = tt.tensor(0, requires_grad=False, **self.factory)
        def _norm(self, x): return x * tt.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)
        def forward(self, x): return self._norm(x.float()).type_as(x) * self.weight + self.bias

    class layerN(_Norm):

        def forward(self, x):  return ff.layer_norm(x, self.weight.shape, self.weight, self.bias, self.eps)


class Vocab:

    @staticmethod
    def random_string(vocab, length): return vocab.backward(tt.randint(0, vocab.size, size=(length,)))

    @staticmethod
    def random_string_excluding(vocab, length, exclude): 
        selector = np.arange(vocab.size)
        mask = np.zeros((len(selector,)), dtype=np.int8)
        mask[exclude]=1
        choice = np.random.choice(np.where(mask==0)[0], size=length, replace=True)
        return vocab.backward(choice)


    @staticmethod
    def random_string_including(vocab, length, include): 
        selector = np.arange(vocab.size)
        mask = np.ones((len(selector,)), dtype=np.int8)
        mask[include]=0
        choice = np.random.choice(np.where(mask==0)[0], size=length, replace=True)
        return vocab.backward(choice)

    @staticmethod
    def random_string_including_starting(vocab, length, include, start): 
        selector = np.arange(vocab.size)
        mask = np.ones((len(selector,)), dtype=np.int8)
        mask[include]=0
        choice = np.hstack(( np.array([start]) , np.random.choice(np.where(mask==0)[0], size=length-1, replace=True)))
        return vocab.backward(choice)

    @staticmethod
    def random_string_excluding_starting(vocab, length, exclude, start): 
        selector = np.arange(vocab.size)
        mask = np.zeros((len(selector,)), dtype=np.int8)
        mask[exclude]=1
        choice = np.hstack(( np.array([start]) , np.random.choice(np.where(mask==0)[0], size=length-1, replace=True)))
        return vocab.backward(choice)
    
    class IntVocab:
        r"""
        maps symbols to tokens
            symbols can be any datatype <--- will be converted to int
            tokens are tt.long or integer type
        """

        # special symbols
        UKN, BOS, EOS, PAD  = -1, -2, -3, -4

        def __init__(self, symbols) -> None:
            self.vocab = {} # value v/s symbol
            self.vocab[self.UKN] =  0
            self.vocab[self.BOS] =  1
            self.vocab[self.EOS] =  2
            self.vocab[self.PAD] =  3
            for i,symbol in enumerate(symbols): self.vocab[int(symbol)] = i + 4 #<-- offset
            self.rvocab = list(self.vocab.keys())
            self.size = len(self.rvocab)

        def __len__(self): return self.size

        
        # inplcae forward
        def forward_(self, symbols, dest):
            for i,symbol in enumerate(symbols): dest[i] = self.vocab[int(symbol)]
        def forward1_(self, symbol, dest, i): dest[i] = self.vocab[int(symbol)]
        
        # forward converts symbol to token
        def forward(self, symbols): return [self.vocab[int(symbol)] for symbol in symbols]
        def forward1(self, symbol):  return self.vocab[int(symbol)]

        # backward converts token to symbol
        def backward(self, tokens): return [self.rvocab[int(token)] for token in tokens]
        def backward1(self, token): return self.rvocab[int(token)]

    class FloatVocab:
        r"""
        maps symbols to tokens
            symbols can be any datatype <--- will be converted to float
            tokens are tt.long or integer type
        """

        # special symbols
        UKN, BOS, EOS, PAD  = -1., -2., -3., -4.

        def __init__(self, symbols) -> None:
            self.vocab = {} # value v/s symbol
            self.vocab[self.UKN] =  0
            self.vocab[self.BOS] =  1
            self.vocab[self.EOS] =  2
            self.vocab[self.PAD] =  3
            for i,symbol in enumerate(symbols): self.vocab[float(symbol)] = i + 4 #<-- offset
            self.rvocab = list(self.vocab.keys())
            self.size = len(self.rvocab)

        def __len__(self): return self.size

        
        # inplcae forward
        def forward_(self, symbols, dest):
            for i,symbol in enumerate(symbols): dest[i] = self.vocab[float(symbol)]
        def forward1_(self, symbol, dest, i): dest[i] = self.vocab[float(symbol)]
        
        # forward converts symbol to token
        def forward(self, symbols): return [self.vocab[float(symbol)] for symbol in symbols]
        def forward1(self, symbol):  return self.vocab[float(symbol)]

        # backward converts token to symbol
        def backward(self, tokens): return [self.rvocab[int(token)] for token in tokens]
        def backward1(self, token): return self.rvocab[int(token)]

    class StrVocab:
        r"""
        maps symbols to tokens
            symbols can be any datatype <--- will be converted to str
            tokens are tt.long or integer type
        """

        # special symbols
        UKN, BOS, EOS, PAD  = "<UKN>", "<BOS>", "<EOS>", "<PAD>"

        def __init__(self, symbols) -> None:
            self.vocab = {} # value v/s symbol
            self.vocab[self.UKN] =  0
            self.vocab[self.BOS] =  1
            self.vocab[self.EOS] =  2
            self.vocab[self.PAD] =  3
            
            for i,symbol in enumerate( symbols ): self.vocab[symbol] = i + 4 #<-- offset
            self.rvocab = list(self.vocab.keys())
            self.size = len(self.rvocab)

        def __len__(self): return self.size

        
        # inplcae forward
        def forward_(self, symbols, dest):
            for i,symbol in enumerate(symbols): dest[i] = self.vocab[symbol]
        def forward1_(self, symbol, dest, i): dest[i] = self.vocab[symbol]
        
        # forward converts symbol to token
        def forward(self, symbols): return [self.vocab[symbol] for symbol in symbols]
        def forward1(self, symbol):  return self.vocab[symbol]

        # backward converts token to symbol
        def backward(self, tokens): return [self.rvocab[int(token)] for token in tokens]
        def backward1(self, token): return self.rvocab[int(token)]

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

